import asyncio
import inspect
import json
from functools import partial
from logging import Logger
from typing import Any
from typing import Callable
from typing import Dict
from typing import ForwardRef
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union
from warnings import warn

from aio_pika import IncomingMessage
from aio_pika import Message
from aio_pika.abc import AbstractIncomingMessage
from aio_pika.abc import AbstractMessage
from anyio.to_thread import run_sync
from pydantic import BaseModel
from pydantic import validate_arguments
from pydantic.typing import evaluate_forwardref

from .abc import AbstractPublisher
from .abc import AbstractRPCClient
from .abc import AbstractSchemeRequirement


class Processor:

    static_param_classes = [Logger, AbstractPublisher, AbstractRPCClient]

    async def __call__(self, *args, **kwargs):
        return await self.__process(*args, **kwargs)

    def __init__(
            self,
            call: Callable,
            input_class: Optional[Type[BaseModel]] = None,
            validate_args: bool = False,
    ):
        self._call = call
        if validate_args:
            self._call = validate_arguments(
                config={
                    'arbitrary_types_allowed': True,
                },
            )(self._call)
        if not self._is_coroutine():
            self.__process = self.__process_sync  # type: ignore
        self._input_class = input_class
        self._signature = self._get_typed_signature()
        self._params = self._get_typed_parameters()
        self._static_params: Iterable[inspect.Parameter] = []
        self._dynamic_params: Iterable[inspect.Parameter] = []
        self._split_static_and_dynamic_params()
        self._cached_static_params: Dict[str, Any] = {}
        if self._input_class is None:
            self._get_data_class()
        self._select_solver()
        self._have_static_params = False

    async def __process(self, *args, **kwargs):
        return await self._call(*args, **kwargs)

    async def __process_sync(self, *args, **kwargs):
        func = partial(self._call, **kwargs)
        return await run_sync(func, *args)

    def cache_static_params(self, component, scheme):
        for param in self._static_params:  # type: inspect.Parameter
            if param.annotation is Logger:
                self._cached_static_params[param.name] = component.log
            elif issubclass(param.annotation, AbstractPublisher):
                self._cached_static_params[param.name] = scheme.publisher(param.default)
            elif issubclass(param.annotation, AbstractRPCClient):
                self._cached_static_params[param.name] = scheme.rpc_client(param.default)
            else:
                raise TypeError("Static param cannot be solved")

    async def solve_requirements(self, settings):
        updated_values = {}
        for param_name, requirement in self._cached_static_params.items():
            if isinstance(requirement, AbstractSchemeRequirement):
                updated_values[param_name] = await requirement.resolve(settings)
        self._cached_static_params.update(updated_values)
        if self._cached_static_params:
            self._call = partial(self._call, **self._cached_static_params)

    def call_sync(self, *args, **kwargs):
        assert not self._is_coroutine()
        return self._call(*args, **kwargs)

    def _static_param_classes_with_subclasses(self):
        all_subclasses = []
        for klass in self.static_param_classes:
            all_subclasses.extend(klass.__subclasses__())
        return self.static_param_classes + all_subclasses

    def _split_static_and_dynamic_params(self):
        for param in self._params:  # type: inspect.Parameter
            if param.annotation in self._static_param_classes_with_subclasses():
                self._static_params.append(param)
                self._have_static_params = True
            else:
                self._dynamic_params.append(param)

    @staticmethod
    def wrap_response(
            result: Union[Dict, BaseModel, Message],
            routing_key: Optional[str] = None,
    ) -> Tuple[Message, Optional[str]]:
        if isinstance(result, Message):
            return result, routing_key
        elif isinstance(result, BaseModel):
            json_encoded = result.json(by_alias=True).encode()
            return Message(json_encoded), routing_key
        elif isinstance(result, dict):
            return Message(json.dumps(result).encode()), routing_key

    async def process(self, message: AbstractIncomingMessage) -> Tuple[Message, Optional[str]]:
        solved_params = self._solve_dependencies(message)
        result = await self(**solved_params)
        wrapped_result = self.wrap_response(result)
        return wrapped_result

    @staticmethod
    def _get_typed_annotation(param: inspect.Parameter, globalns: Dict[str, Any]) -> Any:
        annotation = param.annotation
        if isinstance(annotation, str):
            annotation = ForwardRef(annotation)
            annotation = evaluate_forwardref(annotation, globalns, globalns)
        return annotation

    def _get_typed_signature(self) -> inspect.Signature:
        signature = inspect.signature(self._call)
        globalns = getattr(self._call, "__globals__", {})
        typed_params = [
            inspect.Parameter(
                name=param.name,
                kind=param.kind,
                default=param.default,
                annotation=self._get_typed_annotation(param, globalns),
            )
            for param in signature.parameters.values()
        ]
        typed_signature = inspect.Signature(typed_params)
        return typed_signature

    def _get_typed_parameters(self) -> Iterable[inspect.Parameter]:
        return self._get_typed_signature().parameters.values()

    def _is_coroutine(self) -> bool:
        return asyncio.iscoroutinefunction(self._call)

    def _have_no_annotations(self):
        for param in self._params:
            if param.annotation is not inspect.Parameter.empty:
                return False
        return True

    def _oldstyle_annotations(self):
        params = list(self._params)
        if (
                params[0].annotation is inspect.Parameter.empty
                or params[0].annotation is dict
        ) and issubclass(params[1].annotation, AbstractMessage):
            return True
        return False

    def _get_data_class(self):
        message_class_candidate = None
        if self._input_class is None:
            for param in self._params:
                if issubclass(param.annotation, BaseModel):
                    if message_class_candidate is None:
                        message_class_candidate = param.annotation
                    else:
                        # Two different base classes found. We are not
                        # sure which should be used to parse message.
                        message_class_candidate = None
                        raise AssertionError("Two different data classes are found")
        self._input_class = message_class_candidate
        return self._input_class

    def _solve_dependencies_for_data_class(
        self,
        message: IncomingMessage,
    ) -> Dict[str, Any]:
        solved = {}
        assert self._input_class
        parsed_message = self._input_class.parse_raw(message.body)
        parsed_message_dict = parsed_message.dict(exclude_unset=True)
        for param in self._dynamic_params:  # type: inspect.Parameter
            if param.annotation is IncomingMessage:
                solved[param.name] = message
            elif issubclass(param.annotation, BaseModel):
                solved[param.name] = parsed_message  # type: ignore
            elif param.name in parsed_message_dict:
                solved[param.name] = parsed_message_dict[param.name]
            else:
                raise KeyError(f"Key `{param.name}` cannot be solved by dataclass solver")

        return solved

    def _solve_dependencies(self, message: AbstractIncomingMessage) -> Dict[str, Any]:
        raise NotImplementedError("Dependency solver is not set")

    def _solve_dependencies_for_raw_json(self, message: AbstractIncomingMessage) -> Dict[str, Any]:
        solved = {}
        parsed_message = json.loads(message.body)
        for param in self._dynamic_params:  # type: inspect.Parameter
            if param.annotation is IncomingMessage:
                solved[param.name] = message
            elif param.name in parsed_message:
                solved[param.name] = parsed_message[param.name]
            else:
                raise KeyError(f"Key `{param.name}` cannot be solved by JSON solver")
        return solved

    def _solve_dependencies_oldstyle(self, message: AbstractIncomingMessage) -> Dict[str, Any]:
        solved = {}
        parsed_message = json.loads(message.body)
        for i, param in enumerate(self._params):
            if i == 0:
                solved[param.name] = parsed_message
            elif i == 1:
                solved[param.name] = message
            else:
                raise KeyError("Too much arguments for oldstyle solver")
        return solved

    def _select_solver(self):
        if len(list(self._params)) == 2 and (
                self._have_no_annotations() or self._oldstyle_annotations()
        ):
            warn("Oldstyle injections are deprecated. "
                 "Update your processor definition", DeprecationWarning)
            self._solve_dependencies = self._solve_dependencies_oldstyle
        elif self._input_class:
            self._solve_dependencies = self._solve_dependencies_for_data_class
        else:
            self._solve_dependencies = self._solve_dependencies_for_raw_json



from typing import Dict
from typing import Literal
from typing import Union

from aio_pika import connect_robust
from aio_pika.abc import AbstractConnection

from ...settings import AbstractConnectionParams


connections: Dict[str, AbstractConnection] = {}


def generate_anonymous_connection_name(component_name: str) -> str:
    # TODO
    name = component_name
    return name


async def close_all_connections():
    for connection_name, connection in connections.items():
        await connection.close()


async def connect(
    component_name: str,
    connection_settings: Union[AbstractConnectionParams],
    mode: Literal['r', 'w'],
) -> AbstractConnection:
    if connection_settings.name is None:
        connection_settings.name = generate_anonymous_connection_name(component_name)
    full_connection_name = component_name + '_' + mode
    if full_connection_name not in connections:
        params = connection_settings.get_params_dict()
        connection = await connect_robust(**params)
        connections[full_connection_name] = connection
    return connections[full_connection_name]



from aio_pika.abc import AbstractChannel
from aio_pika.abc import AbstractQueue

from ...factories.core.exchange import declare_exchange
from ...settings import ExchangeParams
from ...settings import QueueParams


async def declare_queue(settings: QueueParams, channel: AbstractChannel) -> AbstractQueue:
    if settings.dead_letter_exchange:
        assert isinstance(settings.dead_letter_exchange, ExchangeParams)
        await declare_exchange(settings.dead_letter_exchange, channel)
    return await channel.declare_queue(**settings.get_params_dict(), timeout=10)



from .consumer import Consumer
from .exceptions import NackMessageError
from .publisher import Publisher
from .service import Service
from .rpc import RPC
from .rpc import RPCClient


__all__ = ['Consumer', 'Publisher', 'Service', 'RPC', 'RPCClient', 'NackMessageError']



from typing import Callable
from typing import Literal
from typing import Mapping
from typing import Optional

from ..abc import AbstractSchemeRequirement
from ..factories import factory_dict
from ..settings import ComponentParamsBaseModel


class SchemeRequirement(AbstractSchemeRequirement):

    def __init__(
        self,
        name: str,
        type_: Literal['publisher', 'consumer', 'service', 'rpc_service', 'rpc_client'],
        params: Optional[ComponentParamsBaseModel] = None,
        processor: Optional[Callable] = None,
    ):
        self.name = name
        self.type_ = type_
        self.params = params
        self.factory = factory_dict[type_]
        self.processor = processor

    async def _resolve(self, settings):
        if self.params:
            return await self.factory(self.params)

        registry: Mapping[str, ComponentParamsBaseModel]

        if self.type_ == 'publisher':
            registry = settings.publishers
        elif self.type_ == 'consumer':
            registry = settings.consumers
        elif self.type_ == 'service':
            registry = settings.services
        elif self.type_ in ('rpc_service', 'rpc_client'):
            registry = settings.rpc_services
        else:
            raise NotImplementedError(f"Unknown component type: `{self.type_}`")

        if self.name not in registry:
            raise KeyError(
                f"{self.type_.title()} `{self.name}` is not described in app settings, "
                f"also params are not provided",
            )
        return await self.factory(registry[self.name])

    async def resolve(self, settings):
        resolved = await self._resolve(settings)
        if self.processor:
            resolved.set_processor(self.processor)
        return resolved

    def set_processor(self, processor: Optional[Callable]):
        assert self.processor is None, (f"Processor for component "
                                        f"requirement `{self.name}` already set")
        self.processor = processor



from json import JSONDecodeError
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Optional

from aio_pika.abc import AbstractIncomingMessage
from aio_pika.abc import AbstractQueue

from mela.components.base import ConsumingComponent
from mela.components.exceptions import NackMessageError
from mela.processor import Processor


class Consumer(ConsumingComponent):

    def __init__(
            self,
            name: str,
            prefetch_count: int = 1,
            timeout: Optional[int] = None,
            no_ack: bool = False,
            exclusive: bool = False,
            consumer_tag: Optional[str] = None,
            requeue_broken_messages: bool = True,
            log_level: str = 'info',
            *,
            queue: Optional[AbstractQueue] = None,
    ):
        super().__init__(name, log_level)
        self._prefetch_count: int = prefetch_count
        self._timeout: Optional[int] = timeout
        self._no_ack: bool = no_ack
        self._exclusive: bool = exclusive
        self._consumer_tag: Optional[str] = consumer_tag
        self._queue: Optional[AbstractQueue] = None
        self.requeue_broken_messages = requeue_broken_messages
        if queue:
            self.set_queue(queue)

        self._callback: Optional[
            Callable[
                [
                    AbstractIncomingMessage,
                ],
                Coroutine[
                    Any,
                    Any,
                    None,
                ],
            ],
        ] = None

    def set_queue(self, queue: AbstractQueue):
        self._queue = queue

    def get_queue_name(self) -> str:
        assert self._queue
        return self._queue.name

    def set_processor(self, processor: Processor):
        self._processor = processor

        async def wrapper(message: AbstractIncomingMessage):
            try:
                await processor.process(message)
            except NackMessageError as e:
                await message.nack(requeue=e.requeue)
                self.log.exception("Message is Nacked:")
            except JSONDecodeError:
                self.log.exception("Message cannot be serialized, so we "
                                   "Nack it with requeue=False")
                await message.nack(requeue=False)
            except Exception:
                await message.nack(requeue=self.requeue_broken_messages)
                self.log.exception("Message is broken:")
            else:
                await message.ack()

        self.set_callback(wrapper)

    def set_callback(self, func: Callable[[AbstractIncomingMessage], Coroutine[Any, Any, None]]):
        self._callback = func

    async def consume(self, **kwargs) -> str:
        assert self._callback is not None, "We can't start without a processor, dude"
        assert self._queue is not None, "Queue is not set"
        consumer_tag = await self._queue.consume(
            callback=self._callback,
            no_ack=self._no_ack,
            exclusive=self._exclusive,
            arguments=kwargs,
            consumer_tag=self._consumer_tag,
            timeout=self._timeout,
        )
        self._consumer_tag = consumer_tag
        return consumer_tag

    async def cancel(self, timeout: Optional[int] = None, nowait: bool = False):
        assert self._consumer_tag
        assert self._queue
        return await self._queue.cancel(self._consumer_tag, timeout, nowait)



from typing import Dict
from typing import Optional
from typing import Union

from aio_pika import Message
from aio_pika.abc import AbstractExchange
from aio_pika.abc import AbstractMessage
from aiormq.abc import ConfirmationFrameType
from pydantic import BaseModel

from ..abc import AbstractPublisher
from ..components.base import Component
from ..processor import Processor


class Publisher(Component, AbstractPublisher):

    def __init__(
            self,
            name: str,
            default_routing_key: str = '',
            default_timeout: int = None,
            log_level: str = 'info',
            *,
            exchange: Optional[AbstractExchange] = None,
    ):
        super().__init__(name, log_level)
        self._default_routing_key = default_routing_key
        self._default_timeout = default_timeout
        self._exchange: Optional[AbstractExchange] = None
        if exchange:
            self.set_exchange(exchange)

    def set_exchange(self, exchange: AbstractExchange):
        assert self._exchange is None, "Exchange already is set"
        self._exchange = exchange

    async def publish_message(
        self,
        message: AbstractMessage,
        routing_key: str = None,
        timeout: int = None,
    ) -> Optional[ConfirmationFrameType]:
        assert self._exchange
        if routing_key is None:
            routing_key = self._default_routing_key
        if timeout is None:
            timeout = self._default_timeout
        return await self._exchange.publish(message, routing_key, timeout=timeout)

    async def publish(
            self,
            message: Union[Dict, BaseModel, AbstractMessage],
            routing_key: Optional[str] = None,
    ):
        message, routing_key = Processor.wrap_response(message, routing_key)
        return await self.publish_message(message, routing_key)



import logging
import sys


root = logging.getLogger()
root.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)



import abc
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import envyaml
from aio_pika import ExchangeType
from aio_pika.abc import AbstractExchange
from aio_pika.abc import AbstractQueue
from pydantic import AmqpDsn
from pydantic import BaseModel
from pydantic import BaseSettings
from pydantic import Extra
from pydantic import Field
from pydantic import PrivateAttr
from pydantic.env_settings import SettingsSourceCallable


def yaml_config_settings_source(settings: 'BaseSettings') -> Dict[str, Any]:
    """
    A simple settings source that loads variables from a YAML file
    at the project's root.
    """
    yaml_config = envyaml.EnvYAML(
        settings.Config.yaml_file_path,  # type: ignore
        include_environment=False,
    )
    clear_config = {}
    for key, value in dict(yaml_config).items():
        if '.' not in key:
            clear_config[key] = value
    return clear_config


class AbstractConnectionParams(BaseModel, abc.ABC):
    name: Optional[str] = None

    @abc.abstractmethod
    def get_params_dict(self) -> Dict[str, Any]:
        raise NotImplementedError

    class Config:
        extra = Extra.forbid


class ConnectionParams(AbstractConnectionParams):
    host: str
    port: int
    login: str = Field(alias='username')
    password: str
    virtualhost: str = '/'
    ssl: bool = False
    ssl_options = dict
    timeout: Optional[Union[float, int]] = None
    client_properties: Optional[Dict] = None
    heartbeat: Optional[int] = None

    def get_params_dict(self):
        if self.name:
            self.client_properties = {'connection_name': self.name}
        res = self.dict(exclude={'name'})
        return res


class URLConnectionParams(AbstractConnectionParams):

    url: AmqpDsn

    def get_params_dict(self):
        return self.dict(exclude={'name'})


class ExchangeParams(BaseModel):
    _instance: Optional[AbstractExchange] = PrivateAttr(default=None)

    name: str
    type: ExchangeType = ExchangeType.DIRECT
    durable: bool = True
    auto_delete: bool = False
    internal: bool = False
    passive: bool = False
    arguments: Optional[Dict] = None
    timeout: Optional[Union[float, int]] = None

    def get_params_dict(self):
        return self.dict()


class QueueParams(BaseModel):
    _instance: Optional[AbstractQueue] = PrivateAttr(default=None)

    name: str
    durable: bool = True
    exclusive: bool = False
    passive: bool = False
    auto_delete: bool = False
    dead_letter_exchange: Optional[Union[str, ExchangeParams]] = None
    dead_letter_exchange_type: ExchangeType = ExchangeType.DIRECT
    dead_letter_routing_key: Optional[str] = None

    def solve_dead_letter_exchange(self, exchanges: Dict['str', 'ExchangeParams']):
        if isinstance(self.dead_letter_exchange, str):
            exchange = exchanges.get(self.dead_letter_exchange)
            if not exchange:
                exchange = ExchangeParams(
                    name=self.dead_letter_exchange,
                    type=self.dead_letter_exchange_type,
                )
            self.dead_letter_exchange = exchange

    def get_params_dict(self):
        arguments = {}
        if self.dead_letter_exchange:
            arguments['x-dead-letter-exchange'] = self.dead_letter_exchange.name
            arguments['x-dead-letter-routing-key'] = self.dead_letter_routing_key
        return {
            'name': self.name,
            'durable': self.durable,
            'exclusive': self.exclusive,
            'passive': self.passive,
            'auto_delete': self.auto_delete,
            'arguments': arguments,
        }

    def solve(self, settings: 'Settings'):
        self.solve_dead_letter_exchange(settings.exchanges)


class ComponentParamsBaseModel(BaseModel, abc.ABC):
    name: Optional[str] = None
    log_level: str = 'info'

    @abc.abstractmethod
    def solve(self, settings: 'Settings'):
        raise NotImplementedError

    @abc.abstractmethod
    def get_params_dict(self) -> Dict:
        raise NotImplementedError


class PublisherParams(ComponentParamsBaseModel):
    connection: Union[str, ConnectionParams, URLConnectionParams] = 'default'
    exchange: Union[str, ExchangeParams]
    exchange_type: str = 'direct'  # DEPRECATED will be deleted in v1.2.0
    routing_key: str
    skip_unroutables: bool = False
    queue: Optional[Union[str, QueueParams]] = None
    timeout: Optional[Union[int, float]] = None

    def solve_connection(
            self,
            connections: Dict[str, Union[ConnectionParams, URLConnectionParams]],
    ) -> None:
        if isinstance(self.connection, str):
            if self.connection not in connections:
                raise KeyError(f"Connection `{self.connection}` is not described in config")
            self.connection = connections[self.connection]

    def solve_exchange(self, exchanges: Dict[str, ExchangeParams]) -> None:
        if isinstance(self.exchange, str):
            exchange = exchanges.get(self.exchange)
            if not exchange:
                exchange = ExchangeParams(name=self.exchange, type=self.exchange_type)
                exchanges[self.exchange] = exchange
            else:
                if self.exchange_type != exchange.type:
                    raise AssertionError("Exchange is configured with two different types")
            self.exchange = exchange

    def solve_queue(self, queues: Dict[str, QueueParams]):
        if isinstance(self.queue, str):
            queue = queues.get(self.queue)
            if queue is None:
                queue = QueueParams(name=self.queue)
                queues[self.queue] = queue
            self.queue = queue

    def solve(self, settings: 'Settings', parent_name: Optional[str] = None):
        self.solve_connection(settings.connections)
        self.solve_exchange(settings.exchanges)
        if self.queue:
            self.solve_queue(settings.queues)
            assert isinstance(self.queue, QueueParams)
            self.queue.solve(settings)
        if parent_name and self.name is None:
            self.name = parent_name + '_publisher'

    def get_params_dict(self):
        return {
            'name': self.name,
            'default_timeout': self.timeout,
            'default_routing_key': self.routing_key,
        }


class ConsumerParams(ComponentParamsBaseModel):
    connection: Union[str, ConnectionParams, URLConnectionParams] = 'default'
    exchange: Union[str, ExchangeParams]
    exchange_type: str = 'direct'  # DEPRECATED will be deleted in v1.2.0
    routing_key: str
    queue: Union[str, QueueParams]
    prefetch_count: int = 1
    dead_letter_exchange: Optional[str] = None
    dead_letter_routing_key: Optional[str] = None
    requeue_broken_messages: bool = True

    def solve_connection(
        self,
        connections: Dict[str, Union[ConnectionParams, URLConnectionParams]],
    ) -> None:
        if isinstance(self.connection, str):
            if self.connection not in connections:
                raise KeyError(f"Connection `{self.connection}` is not described in config")
            self.connection = connections[self.connection]

    def solve_exchange(self, exchanges: Dict[str, ExchangeParams]) -> None:
        if isinstance(self.exchange, str):
            exchange = exchanges.get(self.exchange)
            if self.exchange_type != 'direct':
                # TODO deprecation warning
                pass
            if not exchange:
                exchange = ExchangeParams(name=self.exchange, type=self.exchange_type)
            else:
                if self.exchange_type != exchange.type:
                    raise AssertionError("Exchange is configured with two different types")
            self.exchange = exchange

    def solve_queue(self, queues: Dict[str, QueueParams]):
        if isinstance(self.queue, str):
            queue = queues.get(self.queue)
            if queue is None:
                queue = QueueParams(
                    name=self.queue,
                    dead_letter_exchange=self.dead_letter_exchange,
                    dead_letter_routing_key=self.dead_letter_routing_key,
                )
                queues[self.queue] = queue
            self.queue = queue

    def solve(self, settings: 'Settings', parent_name: Optional[str] = None):
        self.solve_connection(settings.connections)
        self.solve_exchange(settings.exchanges)
        self.solve_queue(settings.queues)
        assert isinstance(self.queue, QueueParams)
        self.queue.solve(settings)
        if parent_name and self.name is None:
            self.name = parent_name + '_consumer'

    def get_params_dict(self):
        return {
            'name': self.name,
            'prefetch_count': self.prefetch_count,
            'requeue_broken_messages': self.requeue_broken_messages,
            'log_level': self.log_level,
        }


class ServiceParams(ComponentParamsBaseModel):

    consumer: Union[str, ConsumerParams]
    publisher: Union[str, PublisherParams]
    requeue_broken_messages: Optional[bool] = None

    name: Optional[str] = None

    def solve_consumer(self, consumers: Dict[str, ConsumerParams]):
        if isinstance(self.consumer, str):
            if self.consumer not in consumers:
                raise KeyError(f"Consumer `{self.consumer}` is not declared")
            self.consumer = consumers[self.consumer]
        if self.requeue_broken_messages is not None:
            self.consumer.requeue_broken_messages = self.requeue_broken_messages

    def solve_publisher(self, publishers: Dict[str, PublisherParams]):
        if isinstance(self.publisher, str):
            if self.publisher not in publishers:
                raise KeyError(f"Publisher `{self.publisher}` is not declared")
            self.publisher = publishers[self.publisher]

    def solve(self, settings: 'Settings', deep=True):
        self.solve_consumer(settings.consumers)
        self.solve_publisher(settings.publishers)
        assert isinstance(self.consumer, ConsumerParams)
        assert isinstance(self.publisher, PublisherParams)
        if deep:
            self.consumer.solve(settings, parent_name=self.name)
            self.publisher.solve(settings, parent_name=self.name)

    def get_params_dict(self) -> Dict:
        return {
            'log_level': self.log_level,
            'name': self.name,
        }


class RPCParams(ComponentParamsBaseModel):
    connection: Union[str, ConnectionParams, URLConnectionParams] = 'default'
    worker: Optional[ConsumerParams] = None
    response_publisher: Optional[PublisherParams] = None
    request_publisher: Optional[PublisherParams] = None

    dead_letter_exchange: Optional[str] = None
    dead_letter_routing_key: Optional[str] = None

    exchange: Union[str, ExchangeParams]
    routing_key: str
    queue: Union[str, QueueParams]
    response_exchange: Union[str, ExchangeParams]

    def solve_connection(
        self,
        connections: Dict[str, Union[ConnectionParams, URLConnectionParams]],
    ) -> None:
        if isinstance(self.connection, str):
            if self.connection not in connections:
                raise KeyError(f"Connection `{self.connection}` is not described in config")
            self.connection = connections[self.connection]

    def solve_exchanges(self, exchanges: Dict[str, ExchangeParams]) -> None:
        if isinstance(self.exchange, str):
            exchange = exchanges.get(self.exchange)
            if not exchange:
                exchange = ExchangeParams(name=self.exchange)
            self.exchange = exchange
        if isinstance(self.response_exchange, str):
            exchange = exchanges.get(self.response_exchange)
            if not exchange:
                exchange = ExchangeParams(name=self.response_exchange)
            self.response_exchange = exchange

    def solve_queue(self, queues: Dict[str, QueueParams]):
        if isinstance(self.queue, str):
            queue = queues.get(self.queue)
            if queue is None:
                queue = QueueParams(
                    name=self.queue,
                    dead_letter_exchange=self.dead_letter_exchange,
                    dead_letter_routing_key=self.dead_letter_routing_key,
                )
                queues[self.queue] = queue
            self.queue = queue

    def solve(self, settings: 'Settings'):
        self.solve_connection(settings.connections)
        self.solve_exchanges(settings.exchanges)
        self.solve_queue(settings.queues)
        assert self.name
        if self.worker is None:
            self.worker = ConsumerParams(
                name=self.name + '_service',
                connection=self.connection,
                exchange=self.exchange,
                routing_key=self.routing_key,
                queue=self.queue,
            )
        if self.response_publisher is None:
            self.response_publisher = PublisherParams(
                name=self.name + '_response_publisher',
                connection=self.connection,
                exchange=self.response_exchange,
                routing_key='',
            )
        if self.request_publisher is None:
            self.request_publisher = PublisherParams(
                name=self.name + '_request_publisher',
                connection=self.connection,
                exchange=self.exchange,
                routing_key=self.routing_key,
            )

    def get_params_dict(self) -> Dict:
        pass


class Settings(BaseSettings):

    connections: Dict[str, Union[ConnectionParams, URLConnectionParams]] = {}
    services: Dict[str, ServiceParams] = {}
    consumers: Dict[str, ConsumerParams] = {}
    publishers: Dict[str, PublisherParams] = {}
    exchanges: Dict[str, ExchangeParams] = {}
    queues: Dict[str, QueueParams] = {}
    rpc_services: Dict[str, RPCParams] = Field(default_factory=dict, alias='rpc-services')

    def __init__(self, **values: Any):
        super().__init__(**values)
        for connection_name, connection in self.connections.items():
            connection.name = connection_name
        for rpc_name, rpc_config in self.rpc_services.items():
            rpc_config.name = rpc_name
            rpc_config.solve(self)
        for service_name, service_config in self.services.items():
            service_config.name = service_name
            service_config.solve(self, deep=True)
        for consumer_name, consumer_config in self.consumers.items():
            consumer_config.name = consumer_name
            consumer_config.solve(self)
        for publisher_name, publisher_config in self.publishers.items():
            publisher_config.name = publisher_name
            publisher_config.solve(self)

    class Config:
        yaml_file_path = 'application.yml'

        extra = Extra.ignore

        @classmethod
        def customise_sources(
                cls,
                init_settings: SettingsSourceCallable,
                env_settings: SettingsSourceCallable,
                file_secret_settings: SettingsSourceCallable,
        ) -> Tuple[SettingsSourceCallable, ...]:
            return (
                init_settings,
                env_settings,
                yaml_config_settings_source,
                file_secret_settings,
            )



import asyncio
from typing import Optional

from .components.base import Component
from .components.base import ConsumingComponent
from .factories.core.connection import close_all_connections
from .factories.publisher import publisher
from .factories.rpc import client as rpc_client
from .scheme import MelaScheme
from .settings import Settings


class Mela(MelaScheme):

    def __init__(
            self,
            name: str,
            settings_: Optional[Settings] = None,
            loop: Optional[asyncio.AbstractEventLoop] = None,
    ):
        super().__init__(name)
        self._settings: Optional[Settings] = None
        if settings_:
            self.settings = settings_
        if loop is None:
            loop = asyncio.get_event_loop()
        self._loop: Optional[asyncio.AbstractEventLoop] = loop

    def publisher_sync(self, name):
        return self._loop.run_until_complete(self.publisher_instance(name))

    async def publisher_instance(self, name):
        return await publisher(self.settings.publishers[name])

    async def rpc_client_instance(self, name):
        return await rpc_client(self.settings.rpc_services[name])

    @property
    def settings(self):
        assert self._settings, "Mela is not configured"
        return self._settings

    @settings.setter
    def settings(self, value: Settings):
        """
        Set config of entire Mela app. It's possible only if app is not
        running yet. In other case it will raise `RuntimeError`
        """
        self._settings = value

    def run(self, coro=None, loop: Optional[asyncio.AbstractEventLoop] = None):
        if self._settings is None:
            self.settings = Settings()
        if loop is None:
            loop = self._loop
        assert loop
        self._run_in_loop(coro, loop)

    @staticmethod
    async def waiter():
        try:
            # Wait until terminate
            await asyncio.Future()
        finally:
            await close_all_connections()

    def _run_in_loop(self, coro, loop: asyncio.AbstractEventLoop):
        assert self._settings
        for requirement_name, requirement in list(self.requirements.items()):
            instance: Component = loop.run_until_complete(
                requirement.resolve(self._settings),
            )
            if isinstance(instance, ConsumingComponent):
                loop.run_until_complete(instance.prepare_processor(self, self.settings))
                loop.run_until_complete(instance.consume())
        if coro:
            loop.run_until_complete(coro)
        else:
            loop.run_until_complete(self.waiter())

    def register_scheme(self, scheme_: MelaScheme):
        self.merge(scheme_)



from json import JSONDecodeError
from typing import Optional

from aio_pika.abc import AbstractIncomingMessage

from mela.components import Consumer
from mela.components import Publisher
from mela.components.base import ConsumingComponent
from mela.components.exceptions import NackMessageError
from mela.processor import Processor


class Service(ConsumingComponent):

    def __init__(
            self,
            name: str,
            log_level: str = 'info',
            *,
            publisher: Optional[Publisher] = None,
            consumer: Optional[Consumer] = None,
    ):
        super().__init__(name, log_level)
        self._consumer: Optional[Consumer] = None
        self._publisher: Optional[Publisher] = None
        if consumer:
            self.consumer = consumer
        if publisher:
            self.publisher = publisher

    def set_processor(self, processor: Processor):
        self._processor = processor

        async def on_message(message: AbstractIncomingMessage) -> None:
            try:
                outgoing_message, routing_key = await processor.process(message)
                await self.publisher.publish_message(outgoing_message, routing_key=routing_key)
            except NackMessageError as e:
                await message.nack(requeue=e.requeue)
                self.log.exception("Message is Nacked:")
            except JSONDecodeError:
                self.log.exception("Message cannot be serialized, so we "
                                   "Nack it with requeue=False")
                await message.nack(requeue=False)
            except Exception:
                await message.nack(requeue=self.consumer.requeue_broken_messages)
                self.log.exception("Message is broken:")
            else:
                await message.ack()
        self.consumer.set_callback(on_message)

    @property
    def consumer(self) -> Consumer:
        if self._consumer is None:
            raise RuntimeError("Consumer is not set")
        return self._consumer

    @consumer.setter
    def consumer(self, value: Consumer):
        if self._consumer:
            raise RuntimeError("Consumer already set")
        self._consumer = value

    @property
    def publisher(self) -> Publisher:
        if self._publisher is None:
            raise RuntimeError("Publisher is not set")
        return self._publisher

    @publisher.setter
    def publisher(self, value: Publisher):
        if self._publisher:
            raise RuntimeError("Publisher already set")
        self._publisher = value

    async def consume(self, **kwargs) -> str:
        return await self.consumer.consume(**kwargs)

    async def cancel(self, timeout: Optional[int] = None, nowait: bool = False):
        return await self.consumer.cancel(timeout, nowait)



from asyncio import AbstractEventLoop
from asyncio import Future
from asyncio import Lock
from json import JSONDecodeError
from json import loads
from typing import Optional
from typing import Type
from typing import Union
from uuid import uuid4

from aio_pika.abc import AbstractIncomingMessage
from aio_pika.abc import AbstractMessage
from pydantic import BaseModel

from ..abc import AbstractRPCClient
from ..processor import Processor
from . import Consumer
from . import Publisher
from .base import ConsumingComponent
from .exceptions import NackMessageError


class RPC(ConsumingComponent):

    def __init__(
            self,
            name: str,
            log_level: str = 'info',
            *,
            worker: Optional[Consumer] = None,
            response_publisher: Optional[Publisher] = None,
            request_publisher: Optional[Publisher] = None,
    ):
        super().__init__(name, log_level)
        self._worker: Optional[Consumer] = None
        self._response_publisher: Optional[Publisher] = None
        self._request_publisher: Optional[Publisher] = None
        if worker:
            self._worker = worker
        if response_publisher:
            self._response_publisher = response_publisher
        if request_publisher:
            self._request_publisher = request_publisher
        self._client: Optional[RPCClient] = None

    def set_processor(self, processor: Processor):

        async def on_message(message: AbstractIncomingMessage) -> None:
            try:
                outgoing_message, _ = await processor.process(message)
                outgoing_message.correlation_id = message.correlation_id
                await self._response_publisher.publish_message(
                    outgoing_message,
                    routing_key=message.reply_to,
                )
            except NackMessageError as e:
                await message.nack(requeue=e.requeue)
                self.log.exception("Message is Nacked:")
            except JSONDecodeError:
                self.log.exception("Message cannot be serialized, so we "
                                   "Nack it with requeue=False")
                await message.nack(requeue=False)
            except Exception:
                await message.nack(requeue=self._worker.requeue_broken_messages)
                self.log.exception("Message is broken:")
            else:
                await message.ack()
        self._worker.set_callback(on_message)

    @property
    def client(self):
        assert self._client is not None
        return self._client

    @client.setter
    def client(self, value):
        self._client = value

    async def consume(self, **kwargs) -> str:
        return await self._worker.consume(**kwargs)

    async def cancel(self, timeout: Optional[int] = None, nowait: bool = False):
        return await self._worker.cancel(timeout, nowait)


class RPCClient(ConsumingComponent, AbstractRPCClient):

    def __init__(
            self,
            name: str,
            log_level: str = 'info',
            loop: AbstractEventLoop = None,
            *,
            request_publisher: Optional[Publisher] = None,
            response_consumer: Optional[Consumer] = None,
            response_model: Optional[Type[BaseModel]] = None,
    ):
        super().__init__(
            name=name,
            loop=loop,
            log_level=log_level,
        )
        self._request_publisher: Optional[Publisher] = None
        if request_publisher:
            self._request_publisher = request_publisher
        if response_consumer:
            self._response_consumer = response_consumer
        self._response_model = response_model
        self._futures = {}
        self._consuming = Lock()

    @staticmethod
    def _generate_correlation_id():
        return str(uuid4())

    async def call(self, body: Union[AbstractMessage, BaseModel, dict], headers=None):
        assert self._consuming.locked(), "Consumer is not active"
        message, _ = Processor.wrap_response(body)
        message.correlation_id = self._generate_correlation_id()
        message.reply_to = self._response_consumer.get_queue_name()
        if headers is not None:
            assert isinstance(headers, dict)
            message.headers.update(headers)

        future = self.loop.create_future()
        self._futures[message.correlation_id] = future

        await self._request_publisher.publish_message(message)

        return await future

    def _prepare_callback(self):

        async def on_message(message: AbstractIncomingMessage) -> None:
            if message.correlation_id is None:
                raise KeyError("Message without correlation id")

            if self._response_model:
                parsed_response = self._response_model.parse_raw(message.body)
            else:
                parsed_response = loads(message.body)
            future: Future = self._futures.pop(message.correlation_id, None)
            if future is not None:
                future.set_result(parsed_response)

        self._response_consumer.set_callback(on_message)

    async def consume(self, **kwargs) -> str:
        self._prepare_callback()
        await self._consuming.acquire()
        return await self._response_consumer.consume(**kwargs)

    async def cancel(self, timeout: Optional[int] = None, nowait: bool = False):
        self._consuming.release()
        return await self._response_consumer.cancel(timeout, nowait)

    def set_processor(self, processor):
        pass



from typing import Dict

from ..components import Publisher
from ..factories.core.connection import connect
from ..factories.core.exchange import declare_exchange
from ..factories.core.queue import declare_queue
from ..settings import AbstractConnectionParams
from ..settings import ExchangeParams
from ..settings import PublisherParams
from ..settings import QueueParams


publishers: Dict[str, Publisher] = {}


async def publisher(settings: PublisherParams) -> Publisher:
    assert settings.name
    if settings.name not in publishers:
        assert isinstance(settings.connection, AbstractConnectionParams)
        connection = await connect(settings.name, settings.connection, 'w')
        channel = await connection.channel()
        assert isinstance(settings.exchange, ExchangeParams)
        exchange = await declare_exchange(settings.exchange, channel)
        if settings.queue:
            async with connection.channel() as temp_channel:
                assert isinstance(settings.queue, QueueParams)
                queue = await declare_queue(settings.queue, temp_channel)
                await queue.bind(exchange, routing_key=settings.routing_key)
        instance: Publisher = Publisher(**settings.get_params_dict(), exchange=exchange)
        publishers[settings.name] = instance
    return publishers[settings.name]



from typing import Dict

from ..components import Consumer
from ..factories.core.connection import connect
from ..factories.core.exchange import declare_exchange
from ..factories.core.queue import declare_queue
from ..settings import AbstractConnectionParams
from ..settings import ConsumerParams
from ..settings import ExchangeParams
from ..settings import QueueParams


consumers: Dict[str, Consumer] = {}


async def consumer(settings: ConsumerParams) -> Consumer:
    assert settings.name
    if settings.name not in consumers:
        assert isinstance(settings.connection, AbstractConnectionParams)
        connection = await connect(settings.name, settings.connection, 'r')
        channel = await connection.channel()
        await channel.set_qos(settings.prefetch_count)
        assert isinstance(settings.queue, QueueParams)
        queue = await declare_queue(settings.queue, await connection.channel())
        assert isinstance(settings.exchange, ExchangeParams)
        exchange = await declare_exchange(settings.exchange, channel)
        await queue.bind(exchange, routing_key=settings.routing_key)
        instance = Consumer(**settings.get_params_dict(), queue=queue)
        consumers[settings.name] = instance
    return consumers[settings.name]


async def anonymous_consumer(settings: ConsumerParams) -> Consumer:
    assert isinstance(settings.connection, AbstractConnectionParams)
    assert settings.name
    connection = await connect(settings.name, settings.connection, 'r')
    channel = await connection.channel()
    await channel.set_qos(1)
    settings.queue = QueueParams(name="", durable=False, auto_delete=True, exclusive=True)
    queue = await declare_queue(settings.queue, await connection.channel())
    assert isinstance(settings.exchange, ExchangeParams)
    exchange = await declare_exchange(settings.exchange, channel)
    await queue.bind(exchange, routing_key=queue.name)
    instance = Consumer(**settings.get_params_dict(), queue=queue)
    return instance



import abc
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Union

from aio_pika import Message
from aio_pika.abc import AbstractMessage
from aiormq.abc import ConfirmationFrameType
from pydantic import BaseModel

from .components.base import Component
from .settings import Settings


class AbstractPublisher(abc.ABC):

    @abc.abstractmethod
    async def publish_message(
        self,
        message: AbstractMessage,
        routing_key: str = None,
        timeout: int = None,
    ) -> Optional[ConfirmationFrameType]:
        raise NotImplementedError

    async def publish(
            self,
            message: Union[Dict, BaseModel, Message],
            routing_key: Optional[str] = None,
    ) -> Optional[ConfirmationFrameType]:
        raise NotImplementedError


class AbstractRPCClient(abc.ABC):

    @abc.abstractmethod
    async def call(self, body: Union[AbstractMessage, BaseModel, dict]) -> Union[BaseModel, dict]:
        raise NotImplementedError


class AbstractSchemeRequirement(abc.ABC):

    @abc.abstractmethod
    async def resolve(self, settings: Settings) -> Component:
        raise NotImplementedError

    @abc.abstractmethod
    def set_processor(self, processor: Optional[Callable]):
        raise NotImplementedError



from ..components.service import Service
from ..settings import ConsumerParams
from ..settings import PublisherParams
from ..settings import ServiceParams
from .consumer import consumer
from .publisher import publisher


async def service(settings: ServiceParams) -> 'Service':
    assert isinstance(settings.consumer, ConsumerParams)
    assert isinstance(settings.publisher, PublisherParams)
    assert settings.name
    publisher_instance = await publisher(settings.publisher)
    consumer_instance = await consumer(settings.consumer)
    instance = Service(
        publisher=publisher_instance,
        consumer=consumer_instance,
        **settings.get_params_dict(),
    )
    return instance



from aio_pika.abc import AbstractChannel
from aio_pika.abc import AbstractExchange

from ...settings import ExchangeParams


async def declare_exchange(settings: ExchangeParams, channel: AbstractChannel) -> AbstractExchange:
    return await channel.declare_exchange(**settings.get_params_dict())



from ..components.rpc import RPC
from ..components.rpc import RPCClient
from ..settings import ConsumerParams
from ..settings import PublisherParams
from ..settings import RPCParams
from .consumer import anonymous_consumer
from .consumer import consumer
from .publisher import publisher


async def service(settings: RPCParams) -> 'RPC':
    assert isinstance(settings.worker, ConsumerParams)
    assert isinstance(settings.request_publisher, PublisherParams)
    assert isinstance(settings.response_publisher, PublisherParams)
    assert settings.name
    worker_instance = await consumer(settings.worker)
    response_publisher_instance = await publisher(settings.response_publisher)
    instance = RPC(
        settings.name,
        log_level=settings.log_level,
        worker=worker_instance,
        response_publisher=response_publisher_instance,
    )
    client_instance = await client(settings)
    instance.client = client_instance
    return instance


async def client(settings: RPCParams) -> 'RPCClient':
    assert isinstance(settings.request_publisher, PublisherParams)
    assert isinstance(settings.response_publisher, PublisherParams)
    assert settings.name
    request_publisher_instance = await publisher(settings.request_publisher)

    consumer_params = ConsumerParams(
        name=settings.name + '_client',
        connection=settings.connection,
        exchange=settings.response_exchange,
        routing_key='',
        queue='',
    )

    response_consumer_instance = await anonymous_consumer(consumer_params)
    instance = RPCClient(
        settings.name,
        log_level=settings.log_level,
        request_publisher=request_publisher_instance,
        response_consumer=response_consumer_instance,
    )
    await instance.consume()
    return instance



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.production.txt", "r") as fh:
    install_requires = fh.read()

with open("VERSION", "r") as fh:
    version = fh.read().strip()

setuptools.setup(
    name="mela",
    version=version,
    author="Daniyar Supiyev",
    author_email="undead.thunderbird@gmail.com",
    description="Let's make microservice development fun again!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/alem-research/mela",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Framework :: AsyncIO",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        ""
    ],
    python_requires='>=3.6',
    install_requires=install_requires
)



import abc
import asyncio
import logging
from typing import Optional

from ..log import handler


class Component(abc.ABC):

    def __init__(self, name, log_level=None, loop=None):
        self.name = name
        if loop is None:
            loop = asyncio.get_running_loop()
        self.loop = loop
        self.log = None
        if log_level is not None:
            self.config_logger(log_level)

    def config_logger(self, level: str):
        self.log = logging.getLogger(self.name)
        self.log.addHandler(handler)
        self.log.setLevel(level.upper())


class ConsumingComponent(Component, abc.ABC):

    def __init__(self, name, log_level: str = None, loop: asyncio.AbstractEventLoop = None):
        super().__init__(name, log_level, loop)
        self._processor = None

    @abc.abstractmethod
    def set_processor(self, processor):
        raise NotImplementedError()

    @abc.abstractmethod
    async def consume(self, **kwargs) -> str:
        raise NotImplementedError()

    @abc.abstractmethod
    async def cancel(self, timeout: Optional[int] = None, nowait: bool = False):
        raise NotImplementedError()

    async def prepare_processor(self, scheme, settings):
        if self._processor:
            self._processor.cache_static_params(self, scheme)
            await self._processor.solve_requirements(settings)



from typing import Any
from typing import Dict

from .consumer import consumer
from .publisher import publisher
from .rpc import client as rpc_client
from .rpc import service as rpc_service
from .service import service


factory_dict: Dict[str, Any] = {
    'consumer': consumer,
    'publisher': publisher,
    'service': service,
    'rpc_service': rpc_service,
    'rpc_client': rpc_client,
}

__all__ = ['consumer', 'publisher', 'service', 'rpc_service', 'rpc_client', 'factory_dict']



import warnings
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Type

from pydantic import BaseModel

from ..processor import Processor
from ..settings import ConsumerParams
from ..settings import PublisherParams
from ..settings import RPCParams
from ..settings import ServiceParams
from .requirement import SchemeRequirement


class MelaScheme:

    """
    Scheme is not runnable. It just declare relations between app components.
    """

    def __init__(
            self,
            name: str,
    ):
        self.name: str = name
        self.requirements: Dict['str', SchemeRequirement] = {}

    def register_component_requirement(self, requirement: SchemeRequirement):
        if requirement.name in self.requirements:
            raise KeyError(f"Looks like requirement with name `{requirement.name}` already exists")
        self.requirements[requirement.name] = requirement

    def service(
        self,
        name: str,
        params: Optional[ServiceParams] = None,
        validate_args: bool = False,
        input_class: Type[BaseModel] = None,
    ) -> Callable[[Callable], Callable]:
        requirement = SchemeRequirement(name, 'service', params)

        def decorator(func: Callable) -> Callable:
            processor = Processor(
                func,
                input_class=input_class,
                validate_args=validate_args,
            )
            requirement.set_processor(processor)
            return processor

        self.register_component_requirement(requirement)
        return decorator

    def publisher(
        self,
        name: str,
        params: Optional[PublisherParams] = None,
    ) -> SchemeRequirement:
        requirement = SchemeRequirement(name, 'publisher', params)
        self.register_component_requirement(requirement)
        return requirement

    def consumer(
        self,
        name: str,
        params: ConsumerParams = None,
        validate_args: bool = False,
        input_class: Type[BaseModel] = None,
    ) -> Callable[[Callable], Callable]:
        requirement = SchemeRequirement(name, 'consumer', params)
        self.register_component_requirement(requirement)

        def decorator(func: Callable[..., Any]) -> Callable:
            processor = Processor(
                func,
                input_class=input_class,
                validate_args=validate_args,
            )
            requirement.set_processor(processor)
            return processor
        return decorator

    def rpc_service(
        self,
        name: str,
        params: Optional[RPCParams] = None,
        validate_args: bool = False,
        request_model: Type[BaseModel] = None,
    ):
        requirement = SchemeRequirement(name, 'rpc_service', params)
        self.register_component_requirement(requirement)

        def decorator(func: Callable[..., Any]) -> Callable:
            processor = Processor(
                func,
                input_class=request_model,
                validate_args=validate_args,
            )
            requirement.set_processor(processor)
            return processor
        return decorator

    def rpc_server(self, *args, **kwargs):
        warnings.warn("`rpc_server` component decorator will be "
                      "deprecated soon. Use `rpc_service instead`", DeprecationWarning)
        return self.rpc_service(*args, **kwargs)

    def rpc_client(
        self,
        name: str,
        params: Optional[RPCParams] = None,
    ) -> SchemeRequirement:
        requirement = SchemeRequirement(name, 'rpc_client', params)
        self.register_component_requirement(requirement)
        return requirement

    def merge(self, other: 'MelaScheme') -> 'MelaScheme':
        for requirement in other.requirements.values():
            self.register_component_requirement(requirement)
        return self
